"""Tool data model for communication."""

import inspect
from dataclasses import dataclass
from typing import Any, Optional, Type

from numerous.apps import HTML, Slider
from numerous.utils import MISSING, ToolT


@dataclass(frozen=True)
class ElementDataModel:
    name: str


@dataclass(frozen=True)
class HTMLElementDataModel(ElementDataModel):
    default: str
    type: str = "html"


@dataclass(frozen=True)
class TextFieldDataModel(ElementDataModel):
    default: str
    type: str = "string"


@dataclass(frozen=True)
class NumberFieldDataModel(ElementDataModel):
    default: float
    type: str = "number"


@dataclass(frozen=True)
class SliderElementDataModel(ElementDataModel):
    default: float
    slider_min_value: float
    slider_max_value: float
    type: str = "slider"


@dataclass(frozen=True)
class ActionDataModel(ElementDataModel):
    type: str = "action"


@dataclass(frozen=True)
class ContainerDataModel(ElementDataModel):
    elements: list[ElementDataModel]
    type: str = "container"


@dataclass(frozen=True)
class ToolDataModel:
    name: str
    elements: list[ElementDataModel]


class ToolDataModelError(Exception):
    def __init__(self, cls: Type[ToolT], name: str, annotation: str) -> None:
        super().__init__(
            f"Invalid Tool. Unsupport field: {cls.__name__}.{name}: {annotation}",
        )


class ToolDefinitionError(Exception):
    def __init__(self, cls: Type[ToolT]) -> None:
        self.cls = cls
        super().__init__(
            f"Tool {self.cls.__name__} not defined with the @app decorator",
        )


def dump_data_model(cls: Type[ToolT]) -> ToolDataModel:
    if not getattr(cls, "__numerous_app__", False):
        raise ToolDefinitionError(cls)

    elements = _dump_element_data_models(cls)

    return ToolDataModel(name=cls.__name__, elements=elements)


def _dump_element_from_annotation(
    name: str,
    annotation: Any,  # noqa: ANN401
    default: Any,  # noqa: ANN401
) -> Optional[ElementDataModel]:
    if annotation is str:
        if isinstance(default, str) or default is MISSING:
            default = "" if default is MISSING else default
            return TextFieldDataModel(name=name, default=default)
        if isinstance(default, HTML):
            return HTMLElementDataModel(name=name, default=default.default)
    if annotation is float:
        if isinstance(default, (float, int)) or default is MISSING:
            return NumberFieldDataModel(
                name=name,
                default=float(0.0 if default is MISSING else default),
            )
        if isinstance(default, Slider):
            default_value = (
                default.min_value
                if default.default is MISSING  # type: ignore[comparison-overlap]
                else default.default
            )

            return SliderElementDataModel(
                name=name,
                default=default_value,
                slider_min_value=default.min_value,
                slider_max_value=default.max_value,
            )

    if getattr(annotation, "__container__", False):
        return dump_container_data_model(name, annotation)

    return None


def _dump_element_data_models(cls: Type[ToolT]) -> list[ElementDataModel]:
    elements: list[ElementDataModel] = []
    annotations = cls.__annotations__ if hasattr(cls, "__annotations__") else {}
    for name, annotation in annotations.items():
        default = getattr(cls, name, MISSING)
        if elem := _dump_element_from_annotation(name, annotation, default):
            elements.append(elem)
        else:
            raise ToolDataModelError(cls, name, annotation)

    for name, func in inspect.getmembers(cls, inspect.isfunction):
        if getattr(func, "__action__", False):
            elements.append(ActionDataModel(name))

    return elements


class ContainerDefinitionError(Exception):
    def __init__(self, cls: Type[ToolT]) -> None:
        self.cls = cls
        super().__init__(
            f"Container {self.cls.__name__} not defined with the @container decorator",
        )


def dump_container_data_model(name: str, cls: Type[ToolT]) -> ContainerDataModel:
    if not getattr(cls, "__container__", False):
        raise ContainerDefinitionError(cls)

    elements = _dump_element_data_models(cls)

    return ContainerDataModel(name=name, elements=elements)
